# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides protobuf generation utilities
"""
import shutil
from subprocess import PIPE, Popen
from tempfile import TemporaryDirectory
from collections import defaultdict
from itertools import chain as iterchain
from os.path import dirname, isfile, join as path_join
from os import makedirs

import numpy as np

from .modeling import List, Dict, Enum, _is_namedtuple
from .utils import namedtuple_field_types


_PROTO_SYNTAX = 'syntax = "proto3";'

_package_template = 'package {name};'

_message_template = '''
message {name} {{
{msg_def}
}}'''

_enum_template = '''
enum {name} {{
{enum_def}
}}'''

_service_template = '''
service {name} {{
{service_def}
}}'''

_rpc_template = 'rpc {name} ({msg_in}) returns ({msg_out});'


class NestedTypeError(Exception):
    pass


_type_lookup = {
    int: 'int64',
    str: 'string',
    float: 'double',
    bool: 'bool',
    bytes: 'bytes',
    np.int64: 'int64',
    np.int32: 'int32',
    np.float32: 'float',
    np.float64: 'double',
    np.bool: 'bool'}


def compile_protostr(proto_str, package_name, module_name, out_dir):
    '''Compiles a Python module from a protobuf definition str and returns the module abspath'''
    _assert_protoc()

    with TemporaryDirectory() as tdir:
        protopath = path_join(tdir, package_name, "{}.proto".format(module_name))

        makedirs(dirname(protopath))
        with open(protopath, 'w') as f:
            f.write(proto_str)

        cmd = "protoc --python_out {tdir} --proto_path {tdir} {protopath}".format(tdir=tdir, protopath=protopath).split()
        p = Popen(cmd, stderr=PIPE)
        _, err = p.communicate()
        if p.returncode != 0:
            raise Exception("A failure occurred while generating source code from protobuf: {}".format(err))

        gen_module_name = "{}_pb2.py".format(module_name)
        gen_module_path = path_join(tdir, package_name, gen_module_name)
        if not isfile(gen_module_path):
            raise Exception("An unknown failure occurred while generating Python module {}".format(gen_module_path))

        out_module_path = path_join(out_dir, gen_module_name)
        shutil.copy(gen_module_path, out_module_path)
        return out_module_path


def _assert_protoc():
    '''Raises an Exception if protoc is not found'''
    if shutil.which('protoc') is None:
        raise Exception('The protocol buffers compiler `protoc` was not found. Verify that it is installed and visible in $PATH')


def model2proto(model, package_name):
    '''Converts a Model object to a protobuf schema string'''
    all_types = (iterchain(*(iterchain(_proto_iter(f.input_type),
                                       _proto_iter(f.output_type)) for f in model.methods.values())))

    unique_types = _require_unique(all_types)
    type_names = set(t.__name__ for t in unique_types)

    msg_defs = tuple(_nt2proto(t, type_names) if _is_namedtuple(t) else _enum2proto(t) for t in unique_types)
    service_def = _gen_service(model)
    package_def = _package_template.format(name=package_name)

    defs = (_PROTO_SYNTAX, package_def, service_def) + msg_defs
    return '\n'.join(defs)


def _proto_iter(nt):
    '''Recursively yields all types contained within the NamedTuple relevant to protobuf gen'''
    if _is_namedtuple(nt):
        yield nt

        for t in nt._field_types.values():
            if _is_namedtuple(t):
                yield from _proto_iter(t)
            elif issubclass(t, Enum):
                yield t
            elif issubclass(t, List) or issubclass(t, Dict):
                for tt in t.__args__:
                    yield from _proto_iter(tt)


def _require_unique(types):
    '''Returns a list of unique types. Raises Exception if named types are not uniquely defined'''
    dd = defaultdict(list)
    for t in types:
        dd[t.__name__].append(t)

    for n, l in dd.items():
        if len(l) > 1 and not all(_types_equal(l[0], t) for t in l[1:]):
            raise Exception("Multiple definitions found for type {}: {}".format(n, l))

    return [l[0] for l in dd.values()]


def _types_equal(t1, t2):
    '''Returns True if t1 and t2 types are equal. Can't override __eq__ on NamedTuple unfortunately.'''
    if _is_namedtuple(t1) and _is_namedtuple(t2):
        names_match = t1.__name__ == t2.__name__
        ft1, ft2 = namedtuple_field_types(t1), namedtuple_field_types(t2)
        keys_match = ft1.keys() == ft2.keys()
        values_match = all(_types_equal(v1, v2) for v1, v2 in zip(ft1.values(), ft2.values()))
        return names_match and keys_match and values_match

    if issubclass(t1, Enum) and issubclass(t2, Enum):
        names_match = t1.__name__ == t2.__name__
        enums_match = [(e.name, e.value) for e in t1] == [(e.name, e.value) for e in t2]
        return names_match and enums_match

    else:
        return t1 == t2


def _nt2proto(nt, type_names):
    '''Converts a NamedTuple definition to a protobuf schema str'''
    msg_def = '\n'.join(_field2proto(field, type_, index, type_names)
                        for index, (field, type_) in enumerate(namedtuple_field_types(nt).items(), 1))
    return _message_template.format(name=nt.__name__, msg_def=msg_def)


def _enum2proto(enum):
    '''Converts an Enum definition to a protobuf schema str'''
    names, values = zip(*((e.name, e.value) for e in enum))
    return _gen_enum(names, enum.__name__, values)


def _gen_enum(enums, name, values=None):
    '''Produces an enum protobuf definition. An UNKNOWN enum is added if values are provided without a 0'''
    if values is not None:
        if 0 not in values:
            enums = ('UNKNOWN', ) + tuple(enums)
            values = (0, ) + tuple(values)

        enums, values = zip(*sorted(zip(enums, values), key=lambda x: x[1]))

        # the first enum value must be zero in proto3
        zidx = values.index(0)
        enums = enums[zidx:] + enums[:zidx]
        values = values[zidx:] + values[:zidx]
    else:
        values = range(len(enums))

    enum_def = '\n'.join(("  {} = {};".format(name, idx) for name, idx in zip(enums, values)))
    return _enum_template.format(name=name, enum_def=enum_def)


def _field2proto(name, type_, index, type_names, rjust=None):
    '''Returns a protobuf schema field str from a NamedTuple field'''
    string = None

    if type_ in _type_lookup:
        string = "{} {} = {};".format(_type2proto(type_), name, index)

    elif _is_namedtuple(type_) or issubclass(type_, Enum):
        tn = type_.__name__
        if tn not in type_names:
            raise Exception("Could not build protobuf field using unknown custom type {}".format(tn))
        string = "{} {} = {};".format(tn, name, index)

    elif issubclass(type_, List):
        inner = type_.__args__[0]
        if _is_container(inner):
            raise NestedTypeError("Nested container {} is not yet supported; try using NamedTuple instead".format(type_))
        string = "repeated {}".format(_field2proto(name, inner, index, type_names, 0))

    elif issubclass(type_, Dict):
        k, v = type_.__args__
        if any(map(_is_container, (k, v))):
            raise NestedTypeError("Nested container {} is not yet supported; try using NamedTuple instead".format(type_))
        string = "map<{}, {}> {} = {};".format(_type2proto(k), _type2proto(v), name, index)

    if string is None:
        raise Exception("Could not build protobuf field due to unsupported type {}".format(type_))

    if rjust is None:
        rjust = len(string) + 2

    return string.rjust(rjust, ' ')


def _is_container(t):
    return issubclass(t, Dict) or issubclass(t, List)


def _type2proto(t):
    '''Returns a string corresponding to the protobuf type'''
    if t in _type_lookup:
        return _type_lookup[t]
    elif _is_namedtuple(t) or issubclass(t, Enum):
        return t.__name__
    else:
        raise Exception("Unknown protobuf mapping for type {}".format(t))


def _gen_service(model, name='Model'):
    '''Returns a protobuf service definition string'''
    rpc_comps = ((n, f.input_type.__name__, f.output_type.__name__) for n, f in model.methods.items())
    rpc_defs = '\n'.join(_gen_rpc(*comps) for comps in rpc_comps)
    return _service_template.format(name=name, service_def=rpc_defs)


def _gen_rpc(name, in_, out):
    '''Returns a protobuf rpc definition string'''
    rpc = _rpc_template.format(name=name, msg_in=in_, msg_out=out)
    return rpc.rjust(len(rpc) + 2)



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides a Acumos session for dumping models
"""
import random
import string
import shutil
import json
import fnmatch
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from os import walk, mkdir
from os.path import extsep, exists, abspath, dirname, isdir, isfile, expanduser, relpath, basename, join as path_join
from pathlib import Path
from collections import namedtuple
from glob import glob

from .pickler import AcumosContextManager, dump_model
from .metadata import create_model_meta, Requirements, Options
from .utils import dump_artifact, get_qualname
from .protogen import model2proto, compile_protostr
from .logging import get_logger
from .modeling import Model


logger = get_logger(__name__)

_LICENSE_NAME = 'license.json'
_PYEXT = "{}py".format(extsep)
_PYGLOB = "*{}".format(_PYEXT)

_ServerResponse = namedtuple('ServerResponse', 'status_code reason text')
_DEPR_MSG = ('Usage of `auth_api` is deprecated; provide an onboarding token instead. '
             'See https://pypi.org/project/acumos/ for more information.')


class AcumosSession(object):
    '''
    A session that enables onboarding models to CubeAI
    '''

    def __init__(self):
        pass

    def dump(self, model, name, outdir, requirements=None):
        '''
        Creates a directory located at ``outdir/name`` containing Acumos model artifacts

        Parameters
        ----------
        model : ``ucumos.modeling.Model``
            An Acumos model instance
        name : str
            The name of your model
        outdir : str
            The directory or folder to save your model .zip to
        requirements : ``acumos.metadata.Requirements``, optional
            Additional Python dependencies that you can optionally specify
        '''
        _assert_valid_input(model, requirements)

        with _dump_model(model, name, requirements) as dump_dir:
            outdir = expanduser(outdir)
            dst_path = path_join(outdir, name)
            if isdir(dst_path):
                shutil.rmtree(dst_path)  # 如果目标文件夹已存在，则先删除
            _copy_dir(dump_dir, outdir, name)
            shutil.make_archive(dst_path, 'zip', dst_path)  # 将目标文件夹自动压缩成.zip文件


def _validate_options(options):
    '''Validates and returns an `Options` object'''
    if options is None:
        options = Options()
    elif not isinstance(options, Options):
        raise Exception('The `options` parameter must be of type `acumos.metadata.Options`')
    return options


def _assert_valid_input(model, requirements):
    '''Raises Exception if inputs are invalid'''
    if not isinstance(model, Model):
        raise Exception("Input `model` must be of type {}".format(get_qualname(Model)))

    if requirements is not None and not isinstance(requirements, Requirements):
        raise Exception("Input `requirements` must be of type {}".format(get_qualname(Requirements)))


def _assert_valid_api(param, api, required):
    '''Raises Exception if an api is invalid'''
    if api is None:
        if required:
            raise Exception("AcumosSession.push requires that the API for `{}` be provided".format(param))
    else:
        if not api.startswith('https'):
            logger.warning("Provided `{}` API {} does not begin with 'https'. Your password and token are visible in plaintext!".format(param, api))


@contextmanager
def _dump_model(model, name, requirements=None):
    '''Generates model artifacts and serializes the model'''
    requirements = Requirements() if requirements is None else requirements

    with TemporaryDirectory() as rootdir:

        model_dir = path_join(rootdir, 'model')
        mkdir(model_dir)

        with AcumosContextManager(model_dir) as context:

            with open(context.build_path('model.pkl'), 'wb') as f:
                dump_model(model, f)

            # generate protobuf definition
            proto_pkg = context.parameters['protobuf_package'] = _random_string()
            protostr = model2proto(model, proto_pkg)
            dump_artifact(rootdir, 'model.proto', data=protostr, module=None, mode='w')

            # generate protobuf source code
            module_name = 'model'
            proto_dir = context.create_subdir('scripts', 'acumos_gen', proto_pkg)
            compile_protostr(protostr, proto_pkg, module_name, proto_dir)

            # generate model metadata
            requirements.reqs.update(context.package_names)
            metadata = create_model_meta(model, name, requirements)
            dump_artifact(rootdir, 'metadata.json', data=metadata, module=json, mode='w')

            # bundle user-provided code
            code_dir = context.create_subdir('scripts', 'user_provided')
            Path(code_dir, '.keep').touch()  # may resolve pruning issues when unzipping

            # copy packages and modules
            # pkg_scripts = _gather_package_scripts(requirements.packages)
            # _copy_package_scripts(context, pkg_scripts, code_dir)  # 不再单独仅拷贝.py文件

            # 拷贝整个package目录
            _copy_package_dirs(requirements.packages, code_dir)

            scripts = set(_gather_scripts(context, requirements))
            _copy_scripts(scripts, code_dir)

        shutil.make_archive(model_dir, 'zip', model_dir)  # create zip at same level as parent
        shutil.rmtree(model_dir)  # clean up model directory

        yield rootdir


def _copy_dir(src_dir, outdir, name):
    '''Copies a directory to a new location'''
    dst_path = path_join(outdir, name)
    if isdir(dst_path):
        raise Exception("Model {} has already been dumped".format(dst_path))
    shutil.copytree(src_dir, dst_path)


def _copy_package_scripts(context, scripts, code_dir):
    '''Moves all gathered package scripts to the context code directory'''
    for script_relpath, script_abspath in scripts:
        script_dirname = dirname(script_relpath)
        context_absdir = context.create_subdir(code_dir, script_dirname, exist_ok=True)
        shutil.copy(script_abspath, context_absdir)


def _gather_package_scripts(packages):
    '''Yields (relpath, abspath) tuples of Python scripts from a sequence of packages'''
    for path in packages:
        path = expanduser(path)
        if not isdir(path):
            raise Exception("Path {} is not a directory".format(path))

        for root, dirnames, filenames in walk(path):
            for filename in fnmatch.filter(filenames, '*.py'):
                script_abspath = path_join(root, filename)
                script_relpath = path_join(basename(path), relpath(script_abspath, path))
                yield script_relpath, script_abspath


def _copy_package_dirs(packages, outdir):
    '''拷贝每个package的整个目录至user_provided目录'''
    for path in packages:
        path = expanduser(path)
        if not isdir(path):
            raise Exception("Path {} is not a directory".format(path))

        name = path.split('/')[-1]
        dst_path = path_join(outdir, name)
        shutil.copytree(path, dst_path)


def _gather_scripts(context, reqs):
    '''Yields absolute paths of Python script dependencies'''
    for script in context.scripts:
        yield script.__file__

    for script_path in reqs.scripts:
        script_abspath = abspath(expanduser(script_path))

        if not exists(script_abspath):
            raise Exception("Provided script requirement {} does not exist".format(script_path))

        if isdir(script_abspath):
            globbed_scripts = glob(path_join(script_abspath, _PYGLOB))
            if not globbed_scripts:
                raise Exception("Provided script requirement directory {} does not contain Python scripts".format(script_path))
            else:
                yield from globbed_scripts
        elif isfile(script_abspath) and script_abspath.endswith(_PYEXT):
            yield script_abspath
        else:
            raise Exception("Provided script requirement {} is invalid. See acumos.metadata.Requirements for documentation".format(script_path))


def _copy_scripts(script_paths, code_dir):
    '''Copies individual scripts to a user-provided code directory'''
    for path in script_paths:
        shutil.copy(path, code_dir)


def _random_string(length=32):
    '''Returns a random string containing ascii characters'''
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
from setuptools import setup, find_packages


with open("README.md", "r", encoding='utf-8') as file:
    long_description = file.read()


setup(
    name='ucumos',
    version='0.8.2',
    author='cubeai',
    author_email='cubeai@163.com',
    description='CuybeAI模型打包客户端',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Apache License 2.0',
    packages=find_packages(),
    install_requires=['protobuf',
                      'requests',
                      'numpy',
                      'dill',
                      'appdirs',
                      'filelock'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
    ],
    keywords='ucumos acumos machine learning model modeling artificial intelligence ml ai',
    python_requires='>=3.5, <3.7',
    url='https://github.com/cube-ai/cubeai-python-client.git',

)
